#ifndef IMAGE_H
#define IMAGE_H

#include <stdio.h>
#include <math.h>
#include "graph.h"


class OurImage{
public:
  int width, height;
  unsigned char (**rgb)[3];
  char **mask; //3, sgmented background, 2: segmentd object (initial), 1: object, 0: background 
  unsigned char objT[3], backT[3];
  
  OurImage(void ){
    width = 0;
    height = 0;
    
    rgb = NULL;
    mask = NULL;
  }
  
  ~OurImage(){
    if(rgb != NULL){
      for(int i=0; i<height; i++){
        delete[] rgb[i];
        delete[] mask[i];
      }
      delete[] rgb;
      delete[] mask;
    }
  }

  bool CreateImage(int ImageWidth, int ImageHeight){
    
    width = ImageWidth;
	height = ImageHeight;
    
    rgb = (unsigned char(**)[3])new unsigned char*[height];
    mask = new char*[height];
   
	for(int i=0; i<height; i++){
      rgb[i] = new unsigned char[width][3];
      mask[i] = new char[width];
      for(int j=0; j<width; j++){
     
        rgb[i][j][0] = (unsigned char)(255);
        rgb[i][j][1] = (unsigned char)(255);
        rgb[i][j][2] = (unsigned char)(255);
        
        mask[i][j] = 2;
      }
    }   
 

	return true;
  }

  bool SetPixel( int x, int y, int r, int g, int b )
  {
	  if(x >= 0 && x < width && y >= 0 && y < height)
	  {
		  rgb[y][x][0] = r;
		  rgb[y][x][1] = g;
		  rgb[y][x][2] = b;
	  }

	  return true;
  }
  
  bool readPpm(char szFileName[] ){
    FILE* in = NULL;
    fopen_s(&in, szFileName, "r");

	if( in == NULL )
	{
		return false;
	}
    
    char type[100];
    fscanf(in, "%s", type);
    
    fscanf(in, "%d %d", &width, &height);
    
    int level;
    fscanf(in, "%d", &level);
    
    rgb = (unsigned char(**)[3])new unsigned char*[height];
    mask = new char*[height];
    for(int i=0; i<height; i++){
      rgb[i] = new unsigned char[width][3];
      mask[i] = new char[width];
      for(int j=0; j<width; j++){
        int r, g, b;
        fscanf(in, "%d %d %d", &r, &g, &b);
        rgb[i][j][0] = (unsigned char)(255*r/level);
        rgb[i][j][1] = (unsigned char)(255*g/level);
        rgb[i][j][2] = (unsigned char)(255*b/level);
        
        mask[i][j] = 2;
      }
    }
    
    fclose(in);

	return true;
  }
  
  void computeT(){
    int aveO[3] = {0,0,0};
    int nO = 0;
    
    int aveB[3] = {0,0,0};
    int nB = 0;
    
    for(int i=0; i<height; i++)
      for(int j=0; j<width; j++){
        unsigned char* v = rgb[i][j];
        if(mask[i][j] == 1){
          aveO[0] += v[0];
          aveO[1] += v[1];
          aveO[2] += v[2];
          nO++;
        }
        else if(mask[i][j] == 0){
          aveB[0] += v[0];
          aveB[1] += v[1];
          aveB[2] += v[2];
          nB++;
        }
      }    

    for(int i=0; i<3; i++){
      objT[i] = (unsigned char)(aveO[i]/nO);
      backT[i] = (unsigned char)(aveB[i]/nB);
    }
  }
  
  void writePpm(char name[]){
    FILE* out;
    fopen_s(&out, name, "w");

    fprintf(out, "P3\n%d %d\n255\n", width, height);
    for(int i=0; i<height; i++){
      for(int j=0; j<width; j++){
        if(mask[i][j] == 2 || mask[i][j] == 1){
          unsigned char* v = rgb[i][j];
          fprintf(out, "%d %d %d ", v[0], v[1], v[2]);
        }
        else
          fprintf(out, "255 255 255 ");
      }
      fprintf(out, "\n");
    }
    fclose(out);
  }

  void writePgm(char name[]){
    FILE* out;
    fopen_s(&out, name, "w");

    fprintf(out, "P2\n%d %d\n255\n", width, height);
    for(int i=0; i<height; i++){
      for(int j=0; j<width; j++){
        if(mask[i][j] == 2 || mask[i][j] == 1){
          unsigned char* v = rgb[i][j];
          fprintf(out, "%d ", (v[0]+v[1]+v[2])/3 );
        }
        else
          fprintf(out, "-1 ");
      }
      fprintf(out, "\n");
    }
    fclose(out);
  }
  
  void segment(float lambda){
    int vN = width*height;
    
    Graph::node_id* nodes = new Graph::node_id[vN];
    Graph *g = new Graph();
    
    for(int i=0; i<height; i++)
      for(int j=0; j<width; j++)
        nodes[width*i+j] = g->add_node();
    
    computeT();
    for(int i=0; i<height; i++)
      for(int j=0; j<width; j++){
        float costT, costS;
        if(mask[i][j] == 1){
          costS = 9.0f;
          costT = 0;
        }
        else if(mask[i][j] == 0){
          costS = 0;
          costT = 9.0f;
        }
        else{
          unsigned char* v = rgb[i][j];
          costS = lambda*dist2(v, backT);
          costT = lambda*dist2(v, objT);
        }
        
        g->set_tweights(nodes[width*i+j], costS, costT);
      }
    
    for(int i=0; i<height; i++)
      for(int j=0; j<width; j++){
        unsigned char* v1 = rgb[i][j];
        int n1 = width*i+j;
        
		// y direction
		// 100
        if(i < height-1){
          unsigned char* v2 = rgb[i+1][j];
          int n2 = (i+1)*width + j;
          float cost = (float)(exp(-dist2(v1, v2)*100));
          g->add_edge(nodes[n1], nodes[n2], cost, cost);
        }
        
		// x direction
        if(j < width-1){
          unsigned char* v2 = rgb[i][j+1];
          int n2 = i*width + j+1;
          float cost = (float)(exp(-dist2(v1, v2)*100));
          g->add_edge(nodes[n1], nodes[n2], cost, cost);
        }
        
		// 
        if(j < width-1 && i < height-1){
          unsigned char* v2 = rgb[i][j+1];
          int n2 = i*width + j+1;
          
          unsigned char* v3 = rgb[i+1][j];
          int n3 = (i+1)*width + j;
          
          unsigned char* v4 = rgb[i+1][j+1];
          int n4 = (i+1)*width + j+1;
          
          float cost14 = (float)(exp(-dist2(v1, v4)*100)/1.414);
          g->add_edge(nodes[n1], nodes[n4], cost14, cost14);
          
          float cost23 = (float)(exp(-dist2(v2, v3)*100)/1.414);
          g->add_edge(nodes[n2], nodes[n3], cost23, cost23);
        }
    }
    
    Graph::flowtype flow = g -> maxflow();
    printf("Flow = %f\n", flow);
    
    for(int i=0; i<height; i++){
      for(int j=0; j<width; j++){
        if(mask[i][j] == 0 || mask[i][j] == 1)
          continue;
        
        if(g->what_segment(nodes[i*width+j]) == Graph::SOURCE)
          mask[i][j] = 2;
        else
          mask[i][j] = 3;
      }
    }
    
    delete g;
    delete[] nodes;
  }
  
  static inline float dist2(unsigned char c1[3], unsigned char c2[3]){
    float dr = (float)(c1[0] - c2[0])/255.0f;
    float dg = (float)(c1[1] - c2[2])/255.0f;
    float db = (float)(c1[2] - c2[2])/255.0f;
    return dr*dr + dg*dg + db*db;
  }
};


#endif
